# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

from .base import ConfigStoreBase, ConfigurationNotFound
from .filestore import FileStore
from .redis import RedisStore

__all__ = [
    "ConfigStoreBase",
    "FileStore",
    "RedisStore",
    "ConfigurationNotFound",
]



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import logging as log
import statsd  # type: ignore
import threading
from typing import Optional

STATSD_PREFIX = ""


class StatsdWrapper(threading.local):
    """
    Wrapper around the statsd library that handles our use cases:

    1. Configurable namespace/prefix per service + context
    2. Proper type hints
    """

    context: Optional[str] = None

    def _build_metric_name(self, name: str) -> str:
        return ".".join([STATSD_PREFIX, self.context or "all", name])

    def get_timer(self, name: str) -> statsd.Timer:
        return statsd.Timer(self._build_metric_name(name))

    def get_counter(self, name: str) -> statsd.Counter:
        return statsd.Counter(self._build_metric_name(name))

    def get_gauge(self, name: str) -> statsd.Gauge:
        return statsd.Gauge(self._build_metric_name(name))


STATSD = StatsdWrapper()


class Base:
    """
    Base class for other "minty" classes.

    This base class provides a lazy-loaded "self.logger" and access to the
    global "statsd" object.
    """

    _logger: log.Logger

    @property
    def logger(self) -> log.Logger:
        """Return this object's logger instance, create one if necessary

        :return: A logger object for this instance
        :rtype: logging.Logger
        """
        try:
            self._logger
        except AttributeError:
            self._logger = log.getLogger(self.__class__.__name__)

        return self._logger

    @property
    def statsd(self) -> StatsdWrapper:
        """Return the global statsd instance (wrapper)"""
        return STATSD



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

from .base import ConfigParserBase
from apacheconfig import make_loader  # type: ignore
from threading import BoundedSemaphore
from typing import Any, Dict, cast

max_concurrent_parsers = 1
parser_semaphore = BoundedSemaphore(value=max_concurrent_parsers)


class ApacheConfigParser(ConfigParserBase):
    __slots__ = ["content"]

    def parse(self, content: str) -> dict:
        """Return a dict with the parsed config.

        :return: Parsed configuration
        :rtype: dict
        """
        preparsed_content = super().parse(content)

        if preparsed_content["content"] == "":
            raise ValueError("Cannot parse empty configuration")

        global parser_semaphore
        with make_loader() as loader, parser_semaphore:
            config = cast(Dict[str, Any], loader.loads(content))

        return config



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import json
from .base import ConfigParserBase
from typing import Dict


class JSONConfigParser(ConfigParserBase):
    __slots__ = ["content"]

    def parse(self, content: str) -> Dict:
        """Parse JSON file and return dict.

        :param content: content from JSON file
        :type content: str
        :raises ValueError: if content is empty
        :return: parsed config dict
        :rtype: dict
        """
        preparsed_content = super().parse(content)

        if preparsed_content["content"] == "":
            raise ValueError("Cannot parse empty configuration")

        config = json.loads(preparsed_content["content"])

        return config



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

from pydantic import BaseModel, Field, ValidationError
from pydantic.generics import GenericModel

__all__ = ["IntrospectableObject", "Field", "ValidationError", "GenericModel"]


class IntrospectableObject(BaseModel):
    """Introspectable object based on pydantic"""

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import minty.cqrs
from unittest import mock
from uuid import uuid4


class InfraFactoryMock:
    """Mock for our InfraFactory"""

    def __init__(self, infra):
        self.infra = infra

    def get_infrastructure(self, context, infrastructure_name):
        return self.infra[infrastructure_name]

    def get_config(self, context):
        return {"instance_uuid": "test"}


class MockRepositoryFactory:
    """Mock for our repository factory"""

    def __init__(self, infra_factory):
        self.infrastructure_factory = infra_factory
        self.repositories = {}

    def get_repository(self, name, context, event_service):
        mock_repo = self.repositories[name]
        return mock_repo


class TestBase:
    """Utilities for making tests on our CQRS infrastructure without
    database access
    """

    _cmd_instance_args = None
    session = None

    def load_query_instance(self, domain: object, inframocks: dict = {}):
        """Loads your object with a qry attribute pointing to the query
        instance

        Will set the self.qry property to the loaded Queries() object,
        containing mocks for:
        * Database
        * S3

        Args:
            domain: Domain object to use in this test,
            e.g.: zsnl_domains.case_management
            inframocks: A key-value pair containing mocks for every component
            (e.g { "database": MagicMock })

        """

        mock_session = mock.MagicMock()
        self.qry = self._get_mocked_domain_instance(
            session_mock=mock_session,
            domain=domain,
            instance_type="query",
            inframocks=inframocks,
        )
        self.session = mock_session

    def load_command_instance(self, domain: object, inframocks: dict = {}):
        """Loads your object with a cmd attribute pointing to the command
        instance

        Will set the self.cmd property to the loaded Queries() object,
        containing mocks for:
        * Database
        * S3

        Args:
            domain: Domain object to use in this test,
                    e.g.: zsnl_domains.case_management
            inframocks: A key-value pair containing mocks for every component
                        (e.g { "database": MagicMock })

        """

        self._cmd_instance_args = {"domain": domain, "inframocks": inframocks}

        if not self.session:
            self.session = mock.MagicMock()

        self.cmd = self._get_mocked_domain_instance(
            session_mock=self.session,
            domain=domain,
            instance_type="command",
            inframocks=inframocks,
        )

    def get_cmd(self):
        if not self._cmd_instance_args:
            raise Exception(
                "Please make sure you first load the command_instance"
            )

        self.load_command_instance(**self._cmd_instance_args)
        return self.cmd

    def assert_has_event_name(self, event_name: str, description: str = None):
        list_of_events = [
            ev
            for ev in self.cmd.event_service.event_list
            if ev.event_name == event_name
        ]

        if description:
            assert len(list_of_events) > 0, description
        else:
            assert len(list_of_events) > 0

        return True

    def get_last_sql_query(self):
        (args, kwargs) = self.session.execute.call_args
        query = args[0].compile()

        return query

    def _get_mocked_domain_instance(
        self,
        session_mock,
        domain,
        instance_type,
        inframocks={},
        user_uuid=None,
    ):
        """Generates a mock for a command or query instance"""

        # Some defaults
        mocked_infra = {
            "database": session_mock,
            "s3": mock.MagicMock(),
            **inframocks,
        }

        if not user_uuid:
            user_uuid = uuid4()

        user_info = minty.cqrs.UserInfo(
            user_uuid=user_uuid, permissions={"admin": True}
        )

        user_info = minty.cqrs.UserInfo(
            user_uuid=user_uuid, permissions={"admin": True}
        )

        if instance_type == "command":
            event_service = minty.cqrs.EventService(
                correlation_id=uuid4(),
                domain="domain",
                context="context",
                user_uuid=user_uuid,
                user_info=user_info,
            )
        else:
            event_service = None

        infra = InfraFactoryMock(infra=mocked_infra)
        repo_factory = MockRepositoryFactory(infra_factory=infra)

        for (repo_name, repo_class) in domain.REQUIRED_REPOSITORIES.items():
            repo_factory.repositories[repo_name] = repo_class(
                infrastructure_factory=infra,
                context=None,
                event_service=event_service,
            )

        if instance_type == "query":
            query = domain.get_query_instance(
                repository_factory=repo_factory,
                context=None,
                user_uuid=user_uuid,
            )
            query.user_info = user_info
            return query
        elif instance_type == "command":
            command = domain.get_command_instance(
                repository_factory=repo_factory,
                event_service=event_service,
                context=None,
                user_uuid=user_uuid,
            )
            command.user_info = user_info
            return command



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import collections.abc
import threading
from contextlib import contextmanager
from typing import Dict

_mdc = threading.local()


def set_mdc(key: str, value: str) -> None:
    """Set a value in the MDC

    The MDC's contents are added to each log message.

    :type key: str
    :type value: object
    """
    setattr(_mdc, key, value)


def get_mdc() -> Dict[str, str]:
    """Get an item from the MDC

    :rtype: dict from str to object
    """
    results: Dict[str, str] = {}
    for log_field in dir(_mdc):
        if log_field.startswith("__"):
            continue
        results[log_field] = getattr(_mdc, log_field)
    return results


def clear_mdc() -> None:
    """Clear the MDC"""
    for k in get_mdc():
        delattr(_mdc, k)


@contextmanager
def mdc(**kw_args) -> collections.abc.Generator[None, None, None]:
    """Context manager that adds the specified keys/values to the MDC, and
    clears them afterwards"""
    for k in kw_args:
        v = kw_args[k]
        set_mdc(k, v)

    yield

    clear_mdc()



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import functools
from .cqrs import EventService
from .cqrs.events import EntityChange
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from minty import exceptions
from minty.object import Field, IntrospectableObject
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    TypeVar,
)
from uuid import UUID


class EntityBase(ABC):
    _event_service: Optional[EventService]
    _changes: List[EntityChange]
    _entity_data: Dict[str, Any]

    @property
    @abstractmethod
    def entity_id(self):
        raise NotImplementedError

    @property
    def event_service(self):
        return self._event_service

    @event_service.setter
    def event_service(self, event_service):
        super().__setattr__("_event_service", event_service)

    @property
    def change_log(self):
        try:
            self._changes
        except AttributeError:
            self.clear_change_log()
        return self._changes

    def clear_change_log(self):
        super().__setattr__("_changes", [])

    @property
    def entity_data(self):
        try:
            self._entity_data
        except AttributeError:
            self.clear_entity_data()
        return self._entity_data

    def clear_entity_data(self):
        super().__setattr__("_entity_data", {})

    def __setattr__(self, attr, value):
        try:
            old_value = self.__getattribute__(attr)
            change: EntityChange = {
                "key": attr,
                "old_value": _reflect(old_value),
                "new_value": _reflect(value),
            }

            self.change_log.append(change)
        except AttributeError:
            pass
        super().__setattr__(attr, value)

    def capture_field_values(self, fields: List):
        for field in fields:
            value = self.__getattribute__(field)
            self.entity_data[field] = _reflect(value)


def _reflect(value):
    "Reflect on attribute type and return a JSON serializable value."

    if (
        (value is None)
        or isinstance(value, bool)
        or isinstance(value, int)
        or isinstance(value, float)
    ):
        return value

    if isinstance(value, EntityBase):
        return {
            "type": value.__class__.__name__,
            "entity_id": str(value.entity_id),
        }

    if isinstance(value, Entity):
        return {k: _reflect(v) for k, v in value.entity_dict().items()}
    elif isinstance(value, IntrospectableObject):
        return {k: _reflect(v) for k, v in value.dict().items()}

    if isinstance(value, (set, list, tuple)):
        return [_reflect(i) for i in value]

    if isinstance(value, Enum):
        return value.value

    if isinstance(value, dict):
        return {k: _reflect(v) for k, v in value.items()}

    if isinstance(value, datetime):
        return value.isoformat()

    return str(value)


class ValueObject(IntrospectableObject):
    pass


def _has_changes(changes: List[EntityChange]) -> bool:
    for change in changes:
        new_value = change.get("new_value", None)
        old_value = change.get("old_value", None)
        if new_value != old_value:
            return True

    return False


def _event_decorator(
    event_name: str,
    fire_always: bool,
    extra_fields: Optional[List[str]] = None,
):
    def _build_decorator(wrapped: Callable[..., Any]):
        @functools.wraps(wrapped)
        def wrapper(*args, **kwargs):
            instance = args[0]
            # Clear the changelog when we are not a class method
            if isinstance(instance, Entity):
                instance.entity_changelog = []

            rv = wrapped(*args, **kwargs)

            changelog: List[EntityChange]

            if isinstance(instance, Entity):
                # When used with a regular method, use the regular
                changelog = instance.entity_changelog
                wrapped_entity = instance
            else:
                if not isinstance(rv, Entity):
                    raise exceptions.ConfigurationConflict(
                        "Return value of class method is not an entity."
                    )

                # When called as a class method (to create a new entity)
                # everything is considered a change
                dict_of_changes = rv.entity_dict()

                changelog = []
                for field in dict_of_changes.keys():
                    value = dict_of_changes[field]

                    changelog.append(
                        {
                            "key": field,
                            "old_value": None,
                            "new_value": _reflect(value),
                        }
                    )

                wrapped_entity = rv

            if extra_fields is not None:
                wrapped_entity.capture_field_values(fields=extra_fields)
            if wrapped_entity._event_service:
                if fire_always or _has_changes(changelog):
                    wrapped_entity._event_service.log_event(
                        entity_type=wrapped_entity.__class__.__name__,
                        entity_id=str(wrapped_entity.entity_id),
                        event_name=event_name,
                        changes=changelog,
                        entity_data=wrapped_entity.entity_data,
                    )
            else:
                raise exceptions.ConfigurationConflict(
                    "Wrapped entity does not have _event_service attribute. "
                    "Possible cause: repository did not inject it."
                )

            return wrapped_entity

        return wrapper

    return _build_decorator


class Entity(IntrospectableObject):
    """
    Pydantic based Entity object

    Entity object based on pydantic and the "new way" of creating entities in
    our minty platform. It has the same functionality as the EntityBase object,
    but does not depend on it. Migrationpaths are unsure.
    """

    entity_type: str = Field(
        ...,
        title="Type of Entity",
        description="Unique name of object within the system",
    )
    entity_id: UUID = Field(
        None,
        title="Identifier of Entity",
        description="Globally unique identifier of this entity",
    )
    entity_meta_summary: str = Field(
        None,
        title="Summary of the subject",
        description="Human readable summary of the content of the object",
    )

    entity_relationships: list = Field(
        [],
        title="Names of attributes containing relationships",
        description="Identifies which attributes relate to other entities",
    )

    entity_meta__fields: list = Field(
        ["entity_meta_summary"],
        title="Names of attributes containing meta fields",
        description="Identifies which attributes contain fields for meta info",
    )

    entity_id__fields: list = Field(
        [],
        title="Names of attributes containing meta fields",
        description="Identifies which attributes are moved to entity_id",
    )

    entity_changelog: list = []
    entity_data: dict = {}

    _event_service: Optional[EventService] = None

    def __init__(self, **kwargs):
        """Initialized an Entity object

        Initialized an entity object and calls super().__init__ on the pydantic
        based IntrospectableObject model. Allows the setting of _event_service
        for the event_service engine
        """

        super().__init__(**kwargs)
        if "_event_service" in kwargs:
            object.__setattr__(
                self, "_event_service", kwargs["_event_service"]
            )

    def entity_dict(self):
        """Generates a python dict containing the values of this entity

        Just as the pydantic.dict() method, it returns a dict containing the
        key/values of the attributes of this object.

        The difference is that it will not return private attributes
        (starting with a _) and attributes starting with "entity_". It walks
        recursively over all the related objects.
        """
        rv = {}
        for k, v in dict(self).items():
            if isinstance(v, Entity):
                rv[k] = v.entity_dict()
            elif isinstance(v, IntrospectableObject):
                rv[k] = v.dict(by_alias=True)
            else:
                if k.startswith("_") or k.startswith("entity_"):
                    continue

                rv[k] = v

        return rv

    def __setattr__(self, attr, value):
        old_value = self.__getattribute__(attr)

        rv = super().__setattr__(attr, value)

        if not (attr.startswith("_") or attr.startswith("entity_")):
            change = {
                "key": attr,
                "old_value": _reflect(old_value),
                "new_value": _reflect(value),
            }

            self.entity_changelog.append(change)

        return rv

    def capture_field_values(self, fields: List):
        for field in fields:
            value = self.__getattribute__(field)
            self.entity_data[field] = _reflect(value)

    @staticmethod
    def event(
        name: str,
        fire_always: bool,
        extra_fields: Optional[List[str]] = None,
    ):
        """
        Decorator to defined events on entities.

        Captures entity changes and saves them to the event_service as events.

        If specified, the contents of fields named in `extra_fields` are
        also included in the event.

        If fire_always, the event will be fired always, even when no entity
        changes are recorded.

        When applying this to a class method, the `@classmethod` decorator
        should be the topmost one:

        ```
            @classmethod
            @Entity.event(name="EventNamed", fire_always=True)
            def event_name(cls):
                pass
        ```
        """
        return _event_decorator(
            event_name=name, fire_always=fire_always, extra_fields=extra_fields
        )


E = TypeVar("E", bound=Entity)


class EntityCollection(Generic[E]):
    """Multiple entities."""

    total_results: Optional[int]
    entities: Iterable[E]

    def __init__(
        self, entities: Iterable[E], total_results: Optional[int] = None
    ):
        self.entities = entities
        self.total_results = total_results

    def __iter__(self) -> Iterator[E]:
        return iter(self.entities)


class RedirectResponse:
    """Use or subclass the RedirectResponse class to make sure a
    HTTP found status (302) response is retured, inluding the specified
    redirect location.
    """

    location: str

    def __init__(self, location: str):
        self.location = location



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

from .apache import ApacheConfigParser
from .base import ConfigParserBase
from .json_parser import JSONConfigParser

__all__ = ["ApacheConfigParser", "ConfigParserBase", "JSONConfigParser"]



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

from . import Base
from .infrastructure import InfrastructureFactory
from typing import TYPE_CHECKING, Any, Dict, Type

if TYPE_CHECKING:
    # This causes an import loop if imported outside of type checking
    import minty.cqrs


class RepositoryBase(Base):
    """Base class for repositories using the Minty CQRS/DDD framework"""

    REQUIRED_INFRASTRUCTURE: Dict[str, Any]
    REQUIRED_INFRASTRUCTURE_RO: Dict[str, Any]
    REQUIRED_INFRASTRUCTURE_RW: Dict[str, Any]

    def __init__(
        self,
        infrastructure_factory: InfrastructureFactory,
        context: str,
        event_service,
    ):
        self.infrastructure_factory = infrastructure_factory
        self.context = context
        self.cache: Dict[str, Any] = {}
        self.event_service = event_service

    def _get_infrastructure(self, name):
        return self.infrastructure_factory.get_infrastructure(
            context=self.context, infrastructure_name=name
        )


class Repository(RepositoryBase):
    _for_entity: str
    _events_to_calls: Dict[str, str]

    def save(
        self, user_info: "minty.cqrs.UserInfo" = None, dry_run: bool = False
    ) -> None:
        """Uses the mapping _events_to_calls for calling the repo methods"""

        for ev in self.event_service.get_events_by_type(
            entity_type=self._for_entity
        ):
            if not ev.processed:
                method_to_call = getattr(
                    self, self._events_to_calls[ev.event_name]
                )

                method_to_call(ev, user_info, dry_run)
            if not dry_run:
                ev.processed = True


class RepositoryFactory(Base):
    """Create context-specific "repository" instances for domains"""

    __slots__ = ["infrastructure_factory", "repositories"]

    def __init__(self, infra_factory: InfrastructureFactory):
        """Initialize the repository factory with an infrastructure factory

        :param infra_factory: Infrastructure factory
        :type infra_factory: InfrastructureFactory
        """
        self.infrastructure_factory = infra_factory
        self.repositories: Dict[str, Type[RepositoryBase]] = {}

    def register_repository(self, name: str, repository: Type[RepositoryBase]):
        """Register a repository class with the repository factory.

        :param repository: repository class; will be instantiated when the
            domain code asks for it by name.
        :type repository: object
        """
        self.repositories[name] = repository

    def get_repository(self, name: str, context=None, event_service=None):
        """Retrieve a repository, given a name and optionally a context.

        :param name: name of repository to instantiate
        :type repository: str
        :param context: Context for which to retrieve the repository.
        :type context: object, optional
        :param event_service: Event service instance
        :type event_service: EventService
        :return: An instance of the configured repository, for the specified
            context.
        :rtype: object
        """
        repo_class = self.repositories[name]

        self.logger.debug(
            f"Creating repository of type '{name}' with context "
            + f"'{context}'"
        )

        repo = repo_class(
            context=context,
            infrastructure_factory=self.infrastructure_factory,
            event_service=event_service,
        )

        return repo



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import amqpstorm  # type: ignore
import json
from .cqrs import MiddlewareBase


def AmqpPublisherMiddleware(
    publisher_name: str, infrastructure_name: str = "amqp"
):
    """Return  `_AMQPPublisherClass` instantiated given params.

    :param publisher_name: name of publisher to get from config
    :type publisher_name: str
    :param infrastructure_name: name of amqp infrastructure, defaults to "amqp"
    :type infrastructure_name: str, optional
    :return: _AMQPPublisher middleware
    :rtype: _AMQPPublisher
    """

    class _AMQPPublisher(MiddlewareBase):
        """Publish all events in the event service to AMQP exchange."""

        def __call__(self, func):
            func()

            self.channel = self.infrastructure_factory.get_infrastructure(
                context=self.context, infrastructure_name=infrastructure_name
            )
            config = self.infrastructure_factory.get_config(
                context=self.context
            )
            publish_settings = config[infrastructure_name]["publish_settings"]
            exchange = publish_settings["exchange"]

            timer = self.statsd.get_timer("amqp_write_duration")
            counter = self.statsd.get_counter("amqp_write_number")

            for event in self.event_service.event_list:
                properties = {"content_type": "application/json"}

                user_uuid = str(
                    event.user_info.user_uuid
                    if event.user_info
                    else event.user_uuid
                )

                event_content = json.dumps(
                    {
                        "id": str(event.uuid),
                        "created_date": event.created_date.isoformat(),
                        "correlation_id": str(self.correlation_id),
                        "context": event.context,
                        "domain": event.domain,
                        "user_uuid": user_uuid,
                        "user_info": json.loads(event.user_info.json())
                        if event.user_info
                        else None,
                        "entity_type": event.entity_type,
                        "entity_id": str(event.entity_id),
                        "event_name": event.event_name,
                        "changes": event.changes,
                        "entity_data": event.entity_data,
                    },
                    sort_keys=True,
                )
                message = amqpstorm.Message.create(
                    channel=self.channel,
                    body=event_content,
                    properties=properties,
                )
                formatted_domain = str(event.domain).replace(".", "_")
                routing_key = f"zsnl.v2.{formatted_domain}.{event.entity_type}.{event.event_name}"

                self.logger.debug(
                    f"Publishing event {event.uuid} on {routing_key} - "
                    + str(message.method)
                    + " - "
                    + str(message.properties)
                    + " - EXCHANGE: "
                    + exchange
                )

                with timer.time():
                    message.publish(routing_key=routing_key, exchange=exchange)

                counter.increment()

    return _AMQPPublisher



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import json
import minty
from ..infrastructure import InfrastructureFactory
from ..repository import RepositoryBase, RepositoryFactory
from .events import Event, EventService, event
from abc import ABC, abstractmethod
from copy import copy
from functools import partial
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Protocol,
    Set,
    Type,
    TypedDict,
)
from uuid import UUID

__all__ = [
    "build_command_lookup_table",
    "CommandBase",
    "CQRS",
    "DomainCommandContainer",
    "DomainQueryContainer",
    "event",
    "Event",
    "EventService",
    "MiddlewareBase",
    "SplitCommandBase",
    "SplitQueryBase",
    "QueryBase",
    "UserInfo",
]


class UserInfo:
    "Data structure for user info"
    # Regular class because of this bug in Pydantic:
    # https://github.com/samuelcolvin/pydantic/issues/1536
    # (a defaultdict is used for platform key access)

    user_uuid: UUID
    permissions: Dict[str, bool]

    def __init__(
        self,
        user_uuid: UUID,
        permissions: Dict[str, bool],
        type: Literal["UserInfo"] = "UserInfo",
    ):
        self.type = type
        self.user_uuid = user_uuid
        self.permissions = permissions

    def json(self):
        return json.dumps(
            {
                "type": "UserInfo",
                "user_uuid": str(self.user_uuid),
                "permissions": self.permissions,
            }
        )


class SplitCommandBase(minty.Base, ABC):
    """
    Base class for 'command' parts of domains.

    This version should be used when every command is split out to its own class.
    """

    name: str

    @abstractmethod
    def __call__(self) -> None:
        raise NotImplementedError

    @property
    def cmd(self) -> Any:
        """
        Reference to the DomainCommandContainer instance that this command is part of.

        Can be used to call other commands.
        """
        return self._command_instance

    @cmd.setter
    def cmd(self, value: "DomainCommandContainer"):
        self._command_instance = value

    def __init__(
        self,
        repository_factory: RepositoryFactory,
        context: str,
        user_uuid: UUID,
        event_service: EventService,
        command_instance: "DomainCommandContainer",
    ):
        self.repository_factory = repository_factory
        self.context = context
        self.user_uuid = user_uuid
        self.event_service = event_service
        self.cmd = command_instance

    def get_repository(self, name: str):
        """Initialize and retrieve a repository by name.

        This also passes the event service to the repository constructor.

        :param name: Name of the repository to retrieve
        :return: Initialized repository
        """

        return self.repository_factory.get_repository(
            name=name, context=self.context, event_service=self.event_service
        )

    def _user_has_permission(self, permission: str) -> bool:
        """
        Check if the currently logged in user has the specified permission,
        and return the result.
        """

        # Command instances do not have `user_info` when called from
        # consumers.
        # Once consumers have the full user_info, this entire condition should
        # be removed:
        if not hasattr(self.cmd, "user_info"):
            return True

        return self.cmd.user_info.permissions.get(permission, False)


class SplitQueryBase(minty.Base, ABC):
    """
    Base class for 'query' parts of domains.

    This version should be used when every query is split out to its own class.
    """

    name: str

    @abstractmethod
    def __call__(self) -> Any:
        raise NotImplementedError

    @property
    def qry(self) -> Any:
        """
        Reference to the DomainQueryContainer instance that this query is part of.

        Can be used to call other queries.
        """
        return self._query_instance

    @qry.setter
    def qry(self, value: "DomainQueryContainer"):
        self._query_instance = value

    def __init__(
        self,
        repository_factory: RepositoryFactory,
        context: str,
        user_uuid: UUID,
        query_instance: "DomainQueryContainer",
    ):
        self.repository_factory = repository_factory
        self.context = context
        self.user_uuid = user_uuid
        self.qry = query_instance

    def get_repository(self, name: str):
        """Initialize and retrieve a repository by name.

        This does not pass an event service to the repository constructor, so
        no events can be triggered accidentally.

        :param name: Name of the repository to retrieve
        :return: Initialized repository
        """

        return self.repository_factory.get_repository(
            name=name, context=self.context, event_service=None
        )


_CommandLookupTable = Dict[str, Type[SplitCommandBase]]
_QueryLookupTable = Dict[str, Type[SplitQueryBase]]


def build_command_lookup_table(
    commands: Set[Type[SplitCommandBase]],
) -> _CommandLookupTable:
    """
    Build a command lookup table (keyed on the command's "short name") containing
    the listed commands.
    """

    command_lookup_table: _CommandLookupTable = {}
    for command in commands:
        assert (
            command.name not in command_lookup_table
        ), f"Tried to register second command with name={command.name}"

        command_lookup_table[command.name] = command

    return command_lookup_table


def build_query_lookup_table(
    queries: Set[Type[SplitQueryBase]],
) -> _QueryLookupTable:
    """
    Build a lookup table for queries (keyed on the query's "short name")
    containing the listed queries.
    """

    query_lookup_table: _QueryLookupTable = {}
    for query in queries:
        assert (
            query.name not in query_lookup_table
        ), f"Tried to register second query with name={query.name}"

        query_lookup_table[query.name] = query

    return query_lookup_table


class DomainCommandContainer:
    user_info: UserInfo

    def __init__(
        self,
        repository_factory: RepositoryFactory,
        context: str,
        user_uuid: UUID,
        event_service: EventService,
        command_lookup_table: _CommandLookupTable,
    ):
        self.repository_factory = repository_factory
        self.context = context
        self.user_uuid = user_uuid
        self.event_service = event_service
        self.command_lookup_table = command_lookup_table

    def __getattr__(self, name: str) -> SplitCommandBase:
        try:
            return self.command_lookup_table[name](
                repository_factory=self.repository_factory,
                context=self.context,
                user_uuid=self.user_uuid,
                event_service=self.event_service,
                command_instance=self,
            )
        except KeyError as e:
            raise AttributeError(name) from e


class DomainQueryContainer:
    user_info: UserInfo

    def __init__(
        self,
        repository_factory: RepositoryFactory,
        context: str,
        user_uuid: UUID,
        query_lookup_table: _QueryLookupTable,
    ):
        self.repository_factory = repository_factory
        self.context = context
        self.user_uuid = user_uuid
        self.query_lookup_table = query_lookup_table

    def __getattr__(self, name: str) -> SplitQueryBase:
        try:
            return self.query_lookup_table[name](
                repository_factory=self.repository_factory,
                context=self.context,
                user_uuid=self.user_uuid,
                query_instance=self,
            )
        except KeyError as e:
            raise AttributeError(name) from e


class CommandBase(minty.Base, ABC):
    "Base class for 'command' parts of domains"

    user_info: Optional[UserInfo]

    def __init__(
        self,
        repository_factory: RepositoryFactory,
        context: str,
        user_uuid: UUID,
        event_service: EventService,
    ):
        self.repository_factory = repository_factory
        self.context = context
        self.user_uuid = user_uuid
        self.event_service = event_service

    def get_repository(self, name):
        """Initialize and retrieve a repository by name.

        This also passes the event service to the repository constructor.

        :param name: Name of the repository to retrieve
        :type name: str
        :return: Initialized repository
        :rtype: RepositoryBase
        """

        return self.repository_factory.get_repository(
            name=name, context=self.context, event_service=self.event_service
        )


class QueryBase(minty.Base, ABC):
    "Base class for 'query' parts of domains"

    user_info: Optional[UserInfo]

    def __init__(self, repository_factory, context, user_uuid: UUID):
        self.repository_factory = repository_factory
        self.context = context
        self.user_uuid = user_uuid

    def get_repository(self, name):
        """Initialize and retrieve a repository by name.

        Passes None as the `event_service`, so any calls to state-changing
        commands can raise the right exceptions.

        :param name: Name of the repository to retrieve
        :type name: str
        :return: Initialized repository
        :rtype: RepositoryBase
        """

        return self.repository_factory.get_repository(
            name=name, context=self.context, event_service=None
        )


class QueryMiddleware(minty.Base, ABC):
    __slots__ = [
        "infrastructure_factory",
        "correlation_id",
        "domain",
        "context",
        "user_uuid",
    ]

    def __init__(
        self,
        infrastructure_factory: InfrastructureFactory,
        correlation_id: UUID,
        domain: str,
        context: str,
        user_uuid: UUID,
    ):
        """Initialize the middleware with an infrastructure factory and some
        parameters from the current request.

        :param infrastructure_factory: [description]
        :type infrastructure_factory: InfrastructureFactory
        :param correlation_id: Unique identifier for the current "chain of
            events"
        :type correlation_id: UUID
        :param domain: Domain the command is a part of
        :type domain: str
        :param context: Context the command is being executed in
        :type context: str
        :param user_uuid: UUID of the user executing the command
        :type user_uuid: UUID
        """

        self.infrastructure_factory = infrastructure_factory
        self.correlation_id = correlation_id
        self.domain = domain
        self.context = context
        self.user_uuid = user_uuid

    @abstractmethod
    def __call__(self, func):
        """Call the specified function.

        Override this method (including the defined params) in your middleware
        class. The logic or functionality you want to implement should go
        before and after your call to `func()`.

        Remember to return whatever `func()` returns!

        :param func: function to execute
        :type func: partial function
        """
        pass


class MiddlewareBase(minty.Base, ABC):
    __slots__ = [
        "infrastructure_factory",
        "event_service",
        "correlation_id",
        "domain",
        "context",
        "user_uuid",
    ]

    def __init__(
        self,
        infrastructure_factory: InfrastructureFactory,
        event_service: EventService,
        correlation_id: UUID,
        domain: str,
        context: str,
        user_uuid: UUID,
    ):
        """Initialize the middleware with an infrastructure factory and some
        parameters from the current request.

        :param infrastructure_factory: Infrastructure factory to use
        :type infrastructure_factory: InfrastructureFactory
        :param event_service: Event service instance to use
        :type event_service: EventService
        :param correlation_id: Unique identifier for the current "chain of
            events"
        :type correlation_id: UUID
        :param domain: Domain the command is a part of
        :type domain: str
        :param context: Context the command is being executed in
        :type context: str
        :param user_uuid: UUID of the user executing the command
        :type user_uuid: UUID
        """

        self.infrastructure_factory = infrastructure_factory
        self.event_service = event_service
        self.correlation_id = correlation_id
        self.domain = domain
        self.context = context
        self.user_uuid = user_uuid

    @abstractmethod
    def __call__(self, func):
        """Call instantiated class.

        Override this method (including the defined params) in your middleware
        class. The logic or functionality you want to implement should go
        before and after your call to `func()`.

        :param func: function to execute
        :type func: partial function
        """
        pass


class EventServiceCleanup(MiddlewareBase):
    """Clears the event list when completed."""

    def __call__(self, func):
        try:
            func()
        finally:
            self.event_service.event_list = []


class CommandInfrastructureCleanup(MiddlewareBase):
    """Set current event in infrastructure factory and clear it after the event is finished."""

    def __call__(self, func):
        try:
            func()
        finally:
            self.infrastructure_factory.flush_local_storage()


class StatsdContextMiddleware(MiddlewareBase):
    """Set (and unset) the current 'context' in the per-thread statsd object."""

    def __call__(self, func):
        try:
            minty.STATSD.context = self.context.replace(".", "_")
            func()
        finally:
            minty.STATSD.context = None


class QueryStatsdContextMiddleware(QueryMiddleware):
    """Set (and unset) the current 'context' in the per-thread statsd object"""

    def __call__(self, func):
        try:
            minty.STATSD.context = self.context.replace(".", "_")
            rv = func()
            return rv
        finally:
            minty.STATSD.context = None


class QueryInfrastructureCleanup(QueryMiddleware):
    """Set current event in infrastructure factory and clear it after the query is finished."""

    def __call__(self, func):
        try:
            rv = func()
            return rv
        finally:
            self.infrastructure_factory.flush_local_storage()


class QueryWrapper(minty.Base):
    """Wrapper class for query instances that applies middleware."""

    def __init__(self, query_instance, middleware: list, domain: str):
        self.middleware = middleware
        self.query_instance = query_instance
        self.domain = domain

    def __getattr__(self, attr):
        """Get an attribute on the wrapped class, wrapped by "event" code.

        This event code ensures the command can't return anything, and creates
        an Event instance.

        :param attr: attribute to retrieve
        :type attr: str
        :return: wrapped method
        :rtype: callable
        """
        original_attribute = getattr(self.query_instance, attr)

        if callable(original_attribute):

            def wrapped(*args, **kwargs):
                """Return attribute wrapped in middlewares."""
                wrapped_func = partial(original_attribute, *args, **kwargs)

                for middleware in self.middleware:
                    wrapped_func = partial(middleware, wrapped_func)
                return wrapped_func()

            return wrapped
        else:
            return original_attribute


class CommandWrapper(minty.Base):
    """Wrapper class for command instances. Handles creation of Events."""

    def __init__(self, command_instance, domain: str):
        self.command_instance = command_instance
        self.middlewares: List[MiddlewareBase] = []
        self.domain = domain

    def register_middleware(self, middleware: MiddlewareBase):
        """Register middleware to be wrapped around command.

        From inner to outer layer, the last middleware class to get registered will be
        the outer shell and will be executed first and last.

        :param middleware: middleware to register
        :type middleware: MiddlewareBase
        """
        self.middlewares.append(middleware)

    def __getattr__(self, attr):
        """Get an attribute on the wrapped class, wrapped by "event" code.

        This event code ensures the command can't return anything, and creates
        an Event instance.

        :param attr: attribute to retrieve
        :type attr: str
        :return: wrapped method
        :rtype: callable
        """

        original_attribute = getattr(self.command_instance, attr)
        if callable(original_attribute):

            def wrapped(*args, **kwargs):
                """Return attribute wrapped in middlewares."""
                wrapped_func = partial(original_attribute, **kwargs)
                for middleware in self.middlewares:
                    wrapped_func = partial(middleware, wrapped_func)

                wrapped_func()
                return

            return wrapped
        else:
            return original_attribute


class CQRSDomain(Protocol):
    """
    Type for CQRS domains.

    Every domain should define which repositories it has/needs, a function
    to return a "query" instance, and one to return a "command" instance.
    """

    REQUIRED_REPOSITORIES: Dict[str, Type[RepositoryBase]]

    __name__: str

    @staticmethod
    def get_query_instance(
        repository_factory: RepositoryFactory, context: str, user_uuid: UUID
    ) -> QueryBase:
        ...

    @staticmethod
    def get_command_instance(
        repository_factory: RepositoryFactory,
        context: str,
        user_uuid: UUID,
        event_service: EventService,
    ) -> CommandBase:
        ...


class CQRSDomainDescriptor(TypedDict):
    module: CQRSDomain
    repository_factory: RepositoryFactory
    repository_factory_ro: RepositoryFactory


class CQRS(minty.Base):
    """Keep commands and queries separated.

    CQRS: Command Query Responsibility Separation
    """

    __slots__ = [
        "domains",
        "infrastructure_factory",
        "infrastructure_factory_ro",
        "command_wrapper_middleware",
        "query_middleware",
    ]

    def __init__(
        self,
        domains: Iterable[CQRSDomain],
        infrastructure_factory: InfrastructureFactory,
        command_wrapper_middleware=None,
        query_middleware=None,
    ):
        """Create a new CQRS instance from a list of domains.

        :param domains: iterable returning domains. Domains are classes or
            packages with at least a "REQUIRED_REPOSITORIES" variable defining
            which repositories are necessary to use the domain.
        :type domains: object
        :param infrastructure_factory: Infrastructure factory, created with
            the required configuration, that the repositories can use to
            create infrastructure instances.
        :type infrastructure_factory: InfrastructureFactory
        :param command_wrapper_middleware: Middlewares to be wrapped around
            command.
        :type command_wrapper_middleware: list of MiddlewareBase
        """
        self.domains: Dict[str, CQRSDomainDescriptor] = {}

        if command_wrapper_middleware is None:
            command_wrapper_middleware = []

        if query_middleware is None:
            query_middleware = []

        self.command_wrapper_middleware = command_wrapper_middleware
        self.query_middleware = query_middleware

        self.infrastructure_factory = infrastructure_factory
        self.infrastructure_factory_ro = copy(infrastructure_factory)

        for domain in domains:
            repo_factory = RepositoryFactory(self.infrastructure_factory)
            repo_factory_ro = RepositoryFactory(self.infrastructure_factory_ro)

            for name, repo in domain.REQUIRED_REPOSITORIES.items():
                repo_factory.register_repository(name=name, repository=repo)
                repo_factory_ro.register_repository(name=name, repository=repo)

                for name, infra in repo.REQUIRED_INFRASTRUCTURE.items():
                    self.infrastructure_factory.register_infrastructure(
                        name=name, infrastructure=infra
                    )
                    self.infrastructure_factory_ro.register_infrastructure(
                        name=name, infrastructure=infra
                    )

                for name, infra in getattr(
                    repo, "REQUIRED_INFRASTRUCTURE_RW", {}
                ).items():
                    self.infrastructure_factory.register_infrastructure(
                        name=name, infrastructure=infra
                    )

                for name, infra in getattr(
                    repo, "REQUIRED_INFRASTRUCTURE_RO", {}
                ).items():
                    self.infrastructure_factory_ro.register_infrastructure(
                        name=name, infrastructure=infra
                    )

            self.domains[domain.__name__] = {
                "module": domain,
                "repository_factory": repo_factory,
                "repository_factory_ro": repo_factory_ro,
            }

    def get_query_instance(
        self,
        correlation_id: UUID,
        domain: str,
        context: str,
        user_uuid: UUID,
        user_info: Optional[UserInfo] = None,
    ):
        """
        Instantiate and return the "query" part of the specified domain.

        :param domain: name of the domain to get the query instance for
        :param context: context for this query instance
        :param user_uuid: UUID of the user that's doing queries
        :param user_info: UserInfo object representing the user doing queries
        """

        self.logger.debug(
            f"Creating query instance for domain '{domain}' with context "
            + f"'{context}' for user '{user_uuid}'"
        )

        short_domain = domain.split(".")[-1]

        query_instance = self.domains[domain]["module"].get_query_instance(
            self.domains[domain]["repository_factory_ro"],
            context=context,
            user_uuid=user_uuid,
        )
        # TODO pass user_info in parameters to `get_query_instance`
        query_instance.user_info = user_info

        initialized_middleware = []

        # Outermost layer: cleanup infrastructure when we're done
        for mw in [
            *self.query_middleware,
            QueryInfrastructureCleanup,
            QueryStatsdContextMiddleware,
        ]:
            initialized_middleware.append(
                mw(
                    infrastructure_factory=self.infrastructure_factory_ro,
                    correlation_id=correlation_id,
                    domain=domain,
                    context=context,
                    user_uuid=user_uuid,
                )
            )

        wrapped_query = QueryWrapper(
            query_instance=query_instance,
            middleware=initialized_middleware,
            domain=short_domain,
        )

        return wrapped_query

    def get_command_instance(
        self,
        correlation_id: UUID,
        domain: str,
        context: str,
        user_uuid: UUID,
        user_info: Optional[UserInfo] = None,
    ):
        """
        Instantiate and return the "command" instance of the specified domain.

        Command instance is instantiated with optional layers of middleware to handle
        various functions when executing a command. `InfrastructureStateManager` is
        always registered as the outermost layer.

        :param correlation_id: unique identifier for the current chain of
            events.
        :param domain: name of the domain to get the query instance for
        :param context: context for this command instance
        :param user_uuid: UUID of the user that's going to execute commands
        :param user_info: UserInfo object representing the user executing commands
        """
        self.logger.debug(
            f"Creating command instance for domain '{domain}' with context "
            + f"'{context}' for user '{user_uuid}'"
        )

        short_domain = domain.split(".")[-1]

        event_service = EventService(
            correlation_id=correlation_id,
            domain=domain,
            context=context,
            user_uuid=user_uuid,
            user_info=user_info,
        )

        cmd_instance = self.domains[domain]["module"].get_command_instance(
            self.domains[domain]["repository_factory"],
            context=context,
            user_uuid=user_uuid,
            event_service=event_service,
        )
        # TODO pass user_info in parameters to `get_command_instance`
        cmd_instance.user_info = user_info

        cmd_wrapped = CommandWrapper(
            command_instance=cmd_instance, domain=short_domain
        )

        # Always add some "outer layers" to clean up
        for middleware in [
            *self.command_wrapper_middleware,
            CommandInfrastructureCleanup,
            EventServiceCleanup,
            StatsdContextMiddleware,
        ]:
            initialized_middleware = middleware(
                infrastructure_factory=self.infrastructure_factory,
                correlation_id=correlation_id,
                domain=domain,
                context=context,
                user_uuid=user_uuid,
                event_service=event_service,
            )
            cmd_wrapped.register_middleware(initialized_middleware)

        return cmd_wrapped



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import json
from .exceptions import ValidationError
from functools import wraps
from jsonpointer import JsonPointer  # type: ignore
from jsonschema import Draft7Validator  # type: ignore
from typing import NoReturn, Optional
from uuid import UUID

_format_checker = Draft7Validator.FORMAT_CHECKER


def assert_never(value: NoReturn) -> NoReturn:  # pragma: no cover
    """
    Assert that a point in the code is never reached.

    Using this function allows the type checker to warn about the unreachable
    code when using enums or literals.

    Example:
    ```
        AllowedValues = Literal['a', 'b']

        def do_something(x: AllowedValues) -> bool:
            if x == 'a':
                return True
            elif x == 'b':
                return False
            assert_never(x)
    ```
    """
    assert False, "This code should never be reached, got: {0}".format(value)


@_format_checker.checks("uuid", raises=ValueError)
def is_uuid(instance):
    return UUID(hex=str(instance))


def validate_with(schema_data: Optional[bytes]):
    """Ensure the decorated function is called with valid arguments

    :param schema_data: The JSONSchema to validate against.
        Should be UTF-8-encoded JSON bytes.
    :type schema_data: bytes
    :return: Decorator for validating function arguments against `schema`
    """
    assert schema_data is not None, "No validation JSON found"

    schema = json.loads(schema_data.decode("utf-8"))

    def validity_wrapper(f):
        validator = Draft7Validator(
            schema=schema, format_checker=_format_checker
        )

        @wraps(f)
        def check_validity(*args, **kwargs):
            errors = []
            for error in validator.iter_errors(instance=kwargs):
                error_pointer = JsonPointer.from_parts(error.absolute_path)

                errors.append(
                    {
                        "context": error.validator,
                        "message": error.message,
                        "cause": str(error.cause) if error.cause else None,
                        "property": error_pointer.path,
                    }
                )

            if len(errors):
                raise ValidationError(errors)

            return f(*args, **kwargs)

        return check_validity

    return validity_wrapper



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

from .. import Base
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, TypedDict, Union
from uuid import UUID, uuid4

if TYPE_CHECKING:
    from . import UserInfo


class EntityChange(TypedDict):
    key: str
    old_value: Any
    new_value: Any


def event(name: str, extra_fields: Optional[List[str]] = None):
    """Event decorator to capture entity changes and publish events to the event_service.

    :param name: Name of event to publish
    :type name: str
    :param extra_fields: extra fields to capture publish with event, defaults to None
    :type extra_fields: List, optional
    :return: wrapped_entity
    :rtype: func
    """

    def register_event(func):
        def wrapper(*args, **kwargs):
            func(*args, **kwargs)
            wrapped_entity = args[0]

            if extra_fields is not None:
                wrapped_entity.capture_field_values(fields=extra_fields)

            wrapped_entity.event_service.log_event(
                entity_type=wrapped_entity.__class__.__name__,
                entity_id=wrapped_entity.entity_id,
                event_name=name,
                changes=wrapped_entity.change_log,
                entity_data=wrapped_entity.entity_data,
            )

            wrapped_entity.clear_entity_data()
            wrapped_entity.clear_change_log()

        return wrapper

    return register_event


@dataclass
class Event:
    uuid: UUID
    created_date: datetime

    correlation_id: UUID
    domain: str
    context: str
    user_uuid: UUID

    entity_type: str
    # Moving to "fully UUID" requires some careful versioning and planning, as
    # dependent code may need to be adjusted.
    entity_id: Union[str, UUID]

    event_name: str
    changes: List[EntityChange]
    entity_data: dict
    processed: bool = False
    user_info: Optional["UserInfo"] = None

    def format_changes(self) -> dict:
        """Format changes into more workable format.

        :return: dict with changes
        :rtype: dict
        """
        changes = {}
        for change in self.changes:
            changes[change["key"]] = change["new_value"]
        return changes

    def previous_value(self, key):
        """Returns the "old value" from the changelog for the given attribute

        :return: value containing the old value
        :rtype: str
        """
        values = [
            change["old_value"]
            for change in self.changes
            if change["key"] == key
        ]
        return values[0]

    def new_value(self, key):
        """Returns the "new value" from the changelog for the given attribute

        :return: value containing the old value
        :rtype: str
        """
        values = [
            change["new_value"]
            for change in self.changes
            if change["key"] == key
        ]
        return values[0]


class EventService(Base):
    __slots__ = (
        "event_list",
        "correlation_id",
        "domain",
        "context",
        "user_uuid",
        "user_info",
    )

    def __init__(
        self,
        correlation_id: UUID,
        domain: str,
        context: str,
        user_uuid: UUID,
        user_info: Optional["UserInfo"] = None,
    ):
        self.correlation_id = correlation_id
        self.domain = domain
        self.context = context

        self.user_info: Optional[UserInfo]

        # If user_info is specified, it gets priority
        if user_info:
            self.user_info = user_info
            self.user_uuid = user_info.user_uuid
        else:
            self.user_info = None
            self.user_uuid = user_uuid

        self.event_list: List[Event] = []

    def log_event(
        self,
        entity_type: str,
        entity_id: Union[str, UUID],
        event_name: str,
        changes: List[EntityChange],
        entity_data: Dict[str, Any],
    ):
        """Register a new event with the event serivce

        This will create a new event using the `EventFactory` configured on
        initialization and append it to the `event_list`.

        :param event_name: Name of the event that happened
        :type event_name: str
        :param parameters: Dictionary containing the event parameters (like
            entity state pre-event and post-event)
        :type parameters: dict
        """

        uuid = uuid4()
        created_date = datetime.utcnow()

        event = Event(
            uuid=uuid,
            created_date=created_date,
            correlation_id=self.correlation_id,
            domain=self.domain,
            context=self.context,
            user_uuid=self.user_uuid,
            user_info=self.user_info,
            entity_type=entity_type,
            entity_id=entity_id,
            event_name=event_name,
            changes=changes,
            entity_data=entity_data,
        )

        self.logger.info(f"Created Event: {event}")

        self.event_list.append(event)

    def get_events_by_type(self, entity_type: str) -> List[Event]:
        """Get all events for given `entity_type`.

        :param entity_type: entity_type
        :type entity_type: str
        :return: list of events
        :rtype: typing.List[Event]
        """
        events = [
            ev for ev in self.event_list if ev.entity_type == entity_type
        ]
        return events
